import glob
import os
from typing import Dict, List, Union

import torch

from diffusers.utils import is_safetensors_available
from tqdm import tqdm

if is_safetensors_available():
    import safetensors.torch

from huggingface_hub import snapshot_download

from diffusers import DiffusionPipeline, __version__, UNet2DConditionModel
from diffusers.schedulers.scheduling_utils import SCHEDULER_CONFIG_NAME
from diffusers.utils import CONFIG_NAME, DIFFUSERS_CACHE, ONNX_WEIGHTS_NAME, WEIGHTS_NAME


NUM_INPUT_BLOCKS = 12
NUM_MID_BLOCK = 1
NUM_OUTPUT_BLOCKS = 12
NUM_TOTAL_BLOCKS = NUM_INPUT_BLOCKS + NUM_MID_BLOCK + NUM_OUTPUT_BLOCKS

DIFFUSERS_KEY_PREFIX_TO_WEIGHT_INDEX = {
    "time_embedding.": 0,
    "conv_in.": 0,
    "down_blocks.0.resnets.0": 1,
    "down_blocks.0.attentions.0": 1,
    "down_blocks.0.resnets.1": 2,
    "down_blocks.0.attentions.1": 2,
    "down_blocks.0.downsamplers": 3,
    "down_blocks.1.resnets.0": 4,
    "down_blocks.1.attentions.0": 4,
    "down_blocks.1.resnets.1": 5,
    "down_blocks.1.attentions.1": 5,
    "down_blocks.1.downsamplers": 6,
    "down_blocks.2.resnets.0": 7,
    "down_blocks.2.attentions.0": 7,
    "down_blocks.2.resnets.1": 8,
    "down_blocks.2.attentions.1": 8,
    "down_blocks.2.downsamplers": 9,
    "down_blocks.3.resnets.0": 10,
    "down_blocks.3.resnets.1": 11,
    "mid_block": 12,
    "up_blocks.0.resnets.0": 13,
    "up_blocks.0.resnets.1": 14,
    "up_blocks.0.resnets.2": 15,
    "up_blocks.0.upsamplers.0": 15,
    "up_blocks.1.resnets.0": 16,
    "up_blocks.1.attentions.0": 16,
    "up_blocks.1.resnets.1": 17,
    "up_blocks.1.attentions.1": 17,
    "up_blocks.1.resnets.2": 18,
    "up_blocks.1.upsamplers.0": 18,
    "up_blocks.1.attentions.2": 18,
    "up_blocks.2.resnets.0": 19,
    "up_blocks.2.attentions.0": 19,
    "up_blocks.2.resnets.1": 20,
    "up_blocks.2.attentions.1": 20,
    "up_blocks.2.resnets.2": 21,
    "up_blocks.2.upsamplers.0": 21,
    "up_blocks.2.attentions.2": 21,
    "up_blocks.3.resnets.0": 22,
    "up_blocks.3.attentions.0": 22,
    "up_blocks.3.resnets.1": 23,
    "up_blocks.3.attentions.1": 23,
    "up_blocks.3.resnets.2": 24,
    "up_blocks.3.attentions.2": 24,
    "conv_norm_out.": 24,
    "conv_out.": 24,
}

def dprint(str, flg):
    if flg:
        print(str)

def get_weight_index(key: str) -> int:
    for k, v in DIFFUSERS_KEY_PREFIX_TO_WEIGHT_INDEX.items():
        if key.startswith(k):
            return v
    raise ValueError(f"Unknown unet key: {key}")


def get_block_alpha(block_weights: list, key: str) -> float:
    weight_index = get_weight_index(key)
    return block_weights[weight_index]


class CheckpointMergerPipeline(DiffusionPipeline):
    """
    A class that that supports merging diffusion models based on the discussion here:
    https://github.com/huggingface/diffusers/issues/877

    Example usage:-

    pipe = DiffusionPipeline.from_pretrained("CompVis/stable-diffusion-v1-4", custom_pipeline="checkpoint_merger.py")

    merged_pipe = pipe.merge(["CompVis/stable-diffusion-v1-4","prompthero/openjourney"], interp = 'inv_sigmoid', alpha = 0.8, force = True)

    merged_pipe.to('cuda')

    prompt = "An astronaut riding a unicycle on Mars"

    results = merged_pipe(prompt)

    ## For more details, see the docstring for the merge method.

    """

    def __init__(self):
        self.register_to_config()
        super().__init__()

    def _compare_model_configs(self, dict0, dict1):
        if dict0 == dict1:
            return True
        else:
            config0, meta_keys0 = self._remove_meta_keys(dict0)
            config1, meta_keys1 = self._remove_meta_keys(dict1)
            if config0 == config1:
                print(f"Warning !: Mismatch in keys {meta_keys0} and {meta_keys1}.")
                return True
        return False

    def _remove_meta_keys(self, config_dict: Dict):
        meta_keys = []
        temp_dict = config_dict.copy()
        for key in config_dict.keys():
            if key.startswith("_"):
                temp_dict.pop(key)
                meta_keys.append(key)
        return (temp_dict, meta_keys)

    @torch.no_grad()
    def merge(self, pretrained_model_name_or_path_list: List[Union[str, os.PathLike]], **kwargs):
        """
        Returns a new pipeline object of the class 'DiffusionPipeline' with the merged checkpoints(weights) of the models passed
        in the argument 'pretrained_model_name_or_path_list' as a list.

        Parameters:
        -----------
            pretrained_model_name_or_path_list : A list of valid pretrained model names in the HuggingFace hub or paths to locally stored models in the HuggingFace format.

            **kwargs:
                Supports all the default DiffusionPipeline.get_config_dict kwargs viz..

                cache_dir, resume_download, force_download, proxies, local_files_only, use_auth_token, revision, torch_dtype, device_map.

                alpha - The interpolation parameter. Ranges from 0 to 1.  It affects the ratio in which the checkpoints are merged. A 0.8 alpha
                    would mean that the first model checkpoints would affect the final result far less than an alpha of 0.2

                interp - The interpolation method to use for the merging. Supports "sigmoid", "inv_sigmoid", "add_diff" and None.
                    Passing None uses the default interpolation which is weighted sum interpolation. For merging three checkpoints, only "add_diff" is supported.

                force - Whether to ignore mismatch in model_config.json for the current models. Defaults to False.

                block_weights - list of 25 floats for per-block weighting. ref https://rentry.org/Merge_Block_Weight_-china-_v1_Beta#3-basic-theory-explanation

                module_override_alphas - dict of str -> float for per-module alpha overrides eg {'unet': 0.2, 'text_encoder': 0.8}

        """
        # Default kwargs from DiffusionPipeline
        cache_dir = kwargs.pop("cache_dir", DIFFUSERS_CACHE)
        resume_download = kwargs.pop("resume_download", False)
        force_download = kwargs.pop("force_download", False)
        proxies = kwargs.pop("proxies", None)
        local_files_only = kwargs.pop("local_files_only", False)
        use_auth_token = kwargs.pop("use_auth_token", None)
        revision = kwargs.pop("revision", None)
        torch_dtype = kwargs.pop("torch_dtype", None)
        device_map = kwargs.pop("device_map", None)

        alpha = kwargs.pop("alpha", 0.5)
        interp = kwargs.pop("interp", None)
        block_weights: list[float] = kwargs.pop("block_weights", None)
        module_override_alphas: dict[str, float] = kwargs.pop("module_override_alphas", {})

        print("Received list", pretrained_model_name_or_path_list)
        print(f"Combining with alpha={alpha}, interpolation mode={interp}")
        if block_weights is not None:
            print(f"Merging unet using block weights {block_weights}")

        checkpoint_count = len(pretrained_model_name_or_path_list)
        # Ignore result from model_index_json comparision of the two checkpoints
        force = kwargs.pop("force", False)

        # If less than 2 checkpoints, nothing to merge. If more than 3, not supported for now.
        if checkpoint_count > 3 or checkpoint_count < 2:
            raise ValueError(
                "Received incorrect number of checkpoints to merge. Ensure that either 2 or 3 checkpoints are being"
                " passed."
            )

        print("Received the right number of checkpoints")
        # chkpt0, chkpt1 = pretrained_model_name_or_path_list[0:2]
        # chkpt2 = pretrained_model_name_or_path_list[2] if checkpoint_count == 3 else None

        # Validate that the checkpoints can be merged
        # Step 1: Load the model config and compare the checkpoints. We'll compare the model_index.json first while ignoring the keys starting with '_'
        config_dicts = []
        for pretrained_model_name_or_path in pretrained_model_name_or_path_list:
            print(f"loading DiffusionPipeline from {pretrained_model_name_or_path}...")
            config_dict = DiffusionPipeline.load_config(
                pretrained_model_name_or_path,
                cache_dir=cache_dir,
                resume_download=resume_download,
                force_download=force_download,
                proxies=proxies,
                local_files_only=local_files_only,
                use_auth_token=use_auth_token,
                revision=revision,
            )
            config_dicts.append(config_dict)

        comparison_result = True
        for idx in range(1, len(config_dicts)):
            comparison_result &= self._compare_model_configs(config_dicts[idx - 1], config_dicts[idx])
            if not force and comparison_result is False:
                print(config_dicts[0], config_dicts[1])
                raise ValueError("Incompatible checkpoints. Please check model_index.json for the models.")
        print("Compatible model_index.json files found")
        # Step 2: Basic Validation has succeeded. Let's download the models and save them into our local files.
        cached_folders = []
        for pretrained_model_name_or_path, config_dict in zip(pretrained_model_name_or_path_list, config_dicts):
            print(f"Loading {pretrained_model_name_or_path}...")
            folder_names = [k for k in config_dict.keys() if not k.startswith("_")]
            allow_patterns = [os.path.join(k, "*") for k in folder_names]
            allow_patterns += [
                WEIGHTS_NAME,
                SCHEDULER_CONFIG_NAME,
                CONFIG_NAME,
                ONNX_WEIGHTS_NAME,
                DiffusionPipeline.config_name,
            ]
            requested_pipeline_class = config_dict.get("_class_name")
            user_agent = {"diffusers": __version__, "pipeline_class": requested_pipeline_class}

            cached_folder = (
                pretrained_model_name_or_path
                if os.path.isdir(pretrained_model_name_or_path)
                else snapshot_download(
                    pretrained_model_name_or_path,
                    cache_dir=cache_dir,
                    resume_download=resume_download,
                    proxies=proxies,
                    local_files_only=local_files_only,
                    use_auth_token=use_auth_token,
                    revision=revision,
                    allow_patterns=allow_patterns,
                    user_agent=user_agent,
                )
            )
            print("Cached Folder", cached_folder)
            cached_folders.append(cached_folder)

        # Step 3:-
        # Load the first checkpoint as a diffusion pipeline and modify its module state_dict in place
        final_pipe = DiffusionPipeline.from_pretrained(
            cached_folders[0], torch_dtype=torch_dtype, device_map=device_map
        )
        final_pipe.to(self.device)

        if interp == "sigmoid":
            theta_func = CheckpointMergerPipeline.sigmoid
        elif interp == "inv_sigmoid":
            theta_func = CheckpointMergerPipeline.inv_sigmoid
        elif interp == "add_diff":
            theta_func = CheckpointMergerPipeline.add_difference
        else:
            theta_func = CheckpointMergerPipeline.weighted_sum

        # Find each module's state dict.
        for attr in final_pipe.config.keys():
            if not attr.startswith("_"):
                checkpoint_path_1 = os.path.join(cached_folders[1], attr)
                if os.path.exists(checkpoint_path_1):
                    files = list(
                        (
                            *glob.glob(os.path.join(checkpoint_path_1, "*.safetensors")),
                            *glob.glob(os.path.join(checkpoint_path_1, "*.bin")),
                        )
                    )
                    checkpoint_path_1 = files[0] if len(files) > 0 else None
                if len(cached_folders) < 3:
                    checkpoint_path_2 = None
                else:
                    checkpoint_path_2 = os.path.join(cached_folders[2], attr)
                    if os.path.exists(checkpoint_path_2):
                        files = list(
                            (
                                *glob.glob(os.path.join(checkpoint_path_2, "*.safetensors")),
                                *glob.glob(os.path.join(checkpoint_path_2, "*.bin")),
                            )
                        )
                        checkpoint_path_2 = files[0] if len(files) > 0 else None
                # For an attr if both checkpoint_path_1 and 2 are None, ignore.
                # If atleast one is present, deal with it according to interp method, of course only if the state_dict keys match.
                if checkpoint_path_1 is None and checkpoint_path_2 is None:
                    print(f"Skipping {attr}: not present in 2nd or 3rd model")
                    continue
                try:
                    module = getattr(final_pipe, attr)

                    if isinstance(module, bool):  # ignore requires_safety_checker boolean
                        continue
                    theta_0 = getattr(module, "state_dict")
                    theta_0 = theta_0()

                    update_theta_0 = getattr(module, "load_state_dict")
                    theta_1 = (
                        safetensors.torch.load_file(checkpoint_path_1)
                        if (is_safetensors_available() and checkpoint_path_1.endswith(".safetensors"))
                        else torch.load(checkpoint_path_1, map_location="cpu")
                    )
                    theta_2 = None
                    if checkpoint_path_2:
                        theta_2 = (
                            safetensors.torch.load_file(checkpoint_path_2)
                            if (is_safetensors_available() and checkpoint_path_2.endswith(".safetensors"))
                            else torch.load(checkpoint_path_2, map_location="cpu")
                        )

                    if not theta_0.keys() == theta_1.keys():
                        print(f"Skipping {attr}: key mismatch")
                        continue
                    if theta_2 and not theta_1.keys() == theta_2.keys():
                        print(f"Skipping {attr}:y mismatch")

                except Exception as e:
                    print(f"Skipping {attr} do to an unexpected error: {str(e)}")
                    continue
                print(f"MERGING {attr}")
                module_alpha = alpha
                if block_weights is not None and type(module) is UNet2DConditionModel:
                    print(f" - using block weights {block_weights}")
                elif module_override_alphas.get(attr, None) is not None:
                    print(f" - using override alpha {module_override_alphas[attr]}")
                    module_alpha = module_override_alphas.get(attr)

                for key in theta_0.keys():
                    block_alpha = (
                        module_alpha if (type(module) is not UNet2DConditionModel or block_weights is None)
                        else get_block_alpha(block_weights, key)
                    )
                    if theta_2:
                        theta_0[key] = theta_func(theta_0[key], theta_1[key], theta_2[key], block_alpha)
                    else:
                        theta_0[key] = theta_func(theta_0[key], theta_1[key], None, block_alpha)

                del theta_1
                del theta_2
                update_theta_0(theta_0)

                del theta_0
        return final_pipe

    @staticmethod
    def weighted_sum(theta0, theta1, theta2, alpha):
        return ((1 - alpha) * theta0) + (alpha * theta1)

    # Smoothstep (https://en.wikipedia.org/wiki/Smoothstep)
    @staticmethod
    def sigmoid(theta0, theta1, theta2, alpha):
        alpha = alpha * alpha * (3 - (2 * alpha))
        return theta0 + ((theta1 - theta0) * alpha)

    # Inverse Smoothstep (https://en.wikipedia.org/wiki/Smoothstep)
    @staticmethod
    def inv_sigmoid(theta0, theta1, theta2, alpha):
        import math

        alpha = 0.5 - math.sin(math.asin(1.0 - 2.0 * alpha) / 3.0)
        return theta0 + ((theta1 - theta0) * alpha)

    @staticmethod
    def add_difference(theta0, theta1, theta2, alpha):
        return theta0 + (theta1 - theta2) * (1.0 - alpha)



# adapted from https://colab.research.google.com/github/pcuenca/diffusers-examples/blob/main/notebooks/stable-diffusion-seeds.ipynb#scrollTo=36026528
import json
import os
import argparse
import pathlib
from typing import Optional

import torch
from diffusers.pipelines.stable_diffusion.convert_from_ckpt import load_pipeline_from_original_stable_diffusion_ckpt
from diffusers import StableDiffusionPipeline, DPMSolverMultistepScheduler
from PIL import Image, ImageDraw, ImageFont
from diffusers.pipelines.stable_diffusion import StableDiffusionPipelineOutput
from diffusers.utils import is_xformers_available
from tqdm import tqdm

from huggingface_hub import list_repo_refs


def disable_nsfw_safety_checker(images, **kwargs):
    return images, False


# get_wrapped_text adapted from https://stackoverflow.com/a/67203353
def get_wrapped_text(text: str, font: ImageFont,
                     line_width: int,
                     max_height: int=None, draw: ImageDraw=None,
                     max_fontsize_reduction_iterations=3) -> tuple[list[str], int|None]:
    """
    Wrap `text` so that when it is drawn using `font` it can fit horizontally within the given
    `line_width` pixels.

    If `max_height` is not None, attempt to reduce the font size at most
    `max_fontsize_reduction_iterations` times until it fits in the requested height. Requires
    that `draw` is also set.

    Returns a list of lines, and a reduced font size or None if no font size reduction
    was calculated.
    """

    remaining_fontsize_reduction_iterations = max_fontsize_reduction_iterations + 1
    reduced_font_size = None
    while True:
        lines = []
        words = text.split()
        current_line = None
        while len(words) > 0:
            next_word = words.pop(0)
            current_line_with_next_word = (next_word
                                           if current_line is None
                                           else f'{current_line} {next_word}')
            if font.getlength(current_line_with_next_word) <= line_width:
                current_line = current_line_with_next_word
            else:
                # break here
                if current_line is not None:
                    lines.append(current_line)
                    words.insert(0, next_word)
                    current_line = None
                else:
                    # single word is too long, need to truncate
                    truncated_word = next_word
                    while len(truncated_word) > 0 and font.getlength(truncated_word) > line_width:
                        truncated_word = truncated_word[:-1]
                    if len(truncated_word) > 0:
                        lines.append(truncated_word)
                        words.insert(0, next_word[len(truncated_word):])
                    else:
                        # give up
                        lines.append(next_word)
        if current_line is not None:
            lines.append(current_line)
        bbox = draw.multiline_textbbox((0,0), "\n".join(lines), font=font) # (left, top, right, bottom)
        if max_height is None or remaining_fontsize_reduction_iterations <= 0 or abs(bbox[3]-bbox[1]) <= max_height:
            # fits in box, or we shouldn't try again
            break
        # reduce font to 75% of current size and try again
        reduced_font_size = round(font.size*0.75)
        font = ImageFont.truetype(font.path, reduced_font_size)
        remaining_fontsize_reduction_iterations -= 1

    return lines, reduced_font_size


def make_label_image(label: str, font: ImageFont, width: int, height: int, margins=None):
    if margins is None:
        margins = [10, 10, 10, 10]
    line_spacing = 5  # pixels between lines
    label_image = Image.new('RGB', size=(width, height), color=(247, 247, 247))
    draw = ImageDraw.Draw(label_image)
    wrapped_lines_list, reduced_font_size_or_none = get_wrapped_text(label,
                                                                     font,
                                                                     line_width=width - (margins[0] + margins[2]),
                                                                     max_height=height - (margins[1] + margins[3]), 
                                                                     draw=draw)
    wrapped_text = '\n'.join(wrapped_lines_list)
    possibly_reduced_font = font if reduced_font_size_or_none is None else ImageFont.truetype(font.path, reduced_font_size_or_none)
    text_bbox = draw.multiline_textbbox((width / 2, 0),
                                        text=wrapped_text,
                                        font=possibly_reduced_font,
                                        anchor='ma',
                                        spacing=line_spacing,
                                        )  # anchor = horizontal middle, ascender
    text_bbox_height = text_bbox[3] - text_bbox[1]
    # print("got text_bbox ", text_bbox)
    y_offset = (height - text_bbox_height) / 2
    x_offset = (width - margins[2] - margins[0]) // 2
    draw.multiline_text((margins[0] + x_offset, margins[1] + y_offset),
                        text=wrapped_text, font=possibly_reduced_font, anchor='ma', fill=(64, 48, 32))
    return label_image


def make_image_grid(imgs, num_rows, num_cols, row_labels: list[str], col_labels: list[str]):
    assert len(imgs) == num_rows * num_cols
    assert len(row_labels) == num_rows
    assert len(col_labels) == num_cols

    print("allocating image...")

    w, h = imgs[0].size
    margin = int(0.125 * w)
    spacing = int(0.125 * w)
    grid = Image.new('RGB', size=(margin + (num_cols + 1) * (w + spacing), margin + (num_rows + 1) * (h + spacing)),
                     color=(255, 255, 255))
    # grid_w, grid_h = grid.size

    # font = ImageFont.load_default()
    font_path = os.path.join(os.path.dirname(__file__), "LibreBaskerville-DpdE.ttf")
    col_font = ImageFont.truetype(font_path, size=32)
    row_font = ImageFont.truetype(font_path, size=48)

    print(f"compositing {len(imgs)} images...")

    for i, l in enumerate(col_labels):
        label_image = make_label_image(l, col_font, w, h, margins=[10, 10, 10, 10])
        grid.paste(label_image, box=(margin + (1 + i) * (w + spacing), margin))

    for i, l in enumerate(row_labels):
        label_image = make_label_image(l, row_font, w, h, margins=[20, 10, 20, 10])
        grid.paste(label_image, box=(margin, margin + (1 + i) * (h + spacing)))

    for i in tqdm(range(len(imgs))):
        img = imgs[i]
        grid.paste(img,
                   box=(margin + (1 + (i % num_cols)) * (w + spacing), margin + (1 + (i // num_cols)) * (h + spacing)))

    return grid


def chunk_list(list, batch_size):
    for i in range(0, len(list), batch_size):
        yield list[i:i + batch_size]


def load_model(repo_id_or_path, prefer_fp16: bool = True, local_files_only: bool=False):
    if os.path.isfile(repo_id_or_path):
        yaml_file = (os.path.splitext(repo_id_or_path)[0] + ".yaml")
        if not os.path.exists(yaml_file):
            yaml_file = None
        return load_pipeline_from_original_stable_diffusion_ckpt(repo_id_or_path, original_config_file=yaml_file)
    elif os.path.isdir(repo_id_or_path):
        return StableDiffusionPipeline.from_pretrained(repo_id_or_path)
    else:
        revision = None  # use default
        if prefer_fp16:
            refs = list_repo_refs(repo_id_or_path)
            fp16_ref = next((r for r in refs.branches if r.name == 'fp16'), None)
            if fp16_ref is not None:
                revision = 'fp16'
        return StableDiffusionPipeline.from_pretrained(repo_id_or_path, revision=revision,
                                                       local_files_only=local_files_only)


def render_row(prompts: list[str],
               negative_prompts: Optional[list[str]],
               seeds: list[int],
               pipeline: StableDiffusionPipeline,
               device: str = None,
               batch_size=1,
               sample_w=512,
               sample_h=512,
               cfg=7.5,
               num_inference_steps=15,  # ddpm++ solver: 15 is typically enough
               disable_nsfw_checker: bool = False,
               ) -> list[Image]:
    # ddpm++
    pipeline.scheduler = DPMSolverMultistepScheduler.from_config(pipeline.scheduler.config,
                                                                 algorithm_type="dpmsolver++")
    if device is None:
        device = 'cuda' if torch.cuda.is_available() else 'mps' if torch.backends.mps.is_available() else 'cpu'
    if device == 'cuda':
        # noinspection PyTypeChecker
        pipeline = pipeline.to(torch.float16)
        if is_xformers_available():
            pipeline.enable_xformers_memory_efficient_attention()
    pipeline = pipeline.to(device)
    if disable_nsfw_checker:
        pipeline.safety_checker = disable_nsfw_safety_checker
    images = []
    negative_prompts = negative_prompts or [""] * len(prompts)
    batches = chunk_list(list(zip(prompts, negative_prompts, seeds)), batch_size)
    progress_bar = tqdm(list(batches))
    for batch in progress_bar:
        batch_prompts, batch_negative_prompts, batch_seeds = zip(*batch)
        print(f" - {batch_prompts}")
        generator_device = 'cpu' if device == 'mps' else device
        manual_seed_generators = [torch.Generator(generator_device).manual_seed(seed) for seed in batch_seeds]
        pipeline_output: StableDiffusionPipelineOutput = pipeline(prompt=list(batch_prompts),
                                                                  negative_prompt=list(batch_negative_prompts),
                                                                  generator=manual_seed_generators,
                                                                  width=sample_w,
                                                                  height=sample_h,
                                                                  num_inference_steps=num_inference_steps,
                                                                  guidance_scale=cfg)
        images += pipeline_output.images

    return images


def merge_models(model_a_repo_id_or_path: str, model_b_repo_id_or_path: str, model_c_repo_id_or_path: Optional[str],
                 alpha=0.5,
                 algorithm: Optional[str] = None, unet_block_weights: Optional[list[float]] = None,
                 per_module_alphas: Optional[dict[str,float]] = None,
                 local_files_only: bool = False) \
        -> StableDiffusionPipeline:
    """
    Merge the two or three given models using the given alpha (for two models: 0.0=100% model a, 1.0=100% model b)
    """

    custom_pipeline = ("checkpoint_merger"
                       if unet_block_weights is None and per_module_alphas is None
                       else os.path.join(pathlib.Path(__file__).parent.resolve(), "checkpoint_merger_mbw.py")
                       )
    pipe = StableDiffusionPipeline.from_pretrained(model_a_repo_id_or_path, local_files_only=local_files_only,
                                                   #custom_pipeline="./checkpoint_merger.py")
                                                    custom_pipeline = custom_pipeline)

    if algorithm is None:
        algorithm = 'add_diff' if model_c_repo_id_or_path is not None else 'weighted_sum'
    if model_c_repo_id_or_path is not None:
        if algorithm != "add_diff":
            raise ValueError("Only add_diff is supported for 3-way merges")
    else:
        if algorithm == "add_diff":
            raise ValueError("add_diff is not supported for 2-way merges")

    force = True
    models = [model_a_repo_id_or_path, model_b_repo_id_or_path]
    if model_c_repo_id_or_path is not None:
        models.append(model_c_repo_id_or_path)
    merged_pipe = pipe.merge(models, local_files_only=local_files_only, interp=algorithm,
                             alpha=alpha, force=force,
                             block_weights=unet_block_weights,
                             module_override_alphas=per_module_alphas)
    del pipe

    return merged_pipe

def explode_merge_config(merge_config: dict) -> list[dict]:

    def get_merge_config_values(key, default):
        v = merge_config.get(key, None)
        if v is None:
            return default
        return v

    merge_alphas = get_merge_config_values('alphas', [0.5])
    merge_algorithms = get_merge_config_values('algorithms', ['weighted_sum'])
    unet_alphas = get_merge_config_values('unet_alphas', [None])
    text_encoder_alphas = get_merge_config_values('text_encoder_alphas', [None])
    unet_block_weights_strs = get_merge_config_values('block_weights', [])
    if len(unet_block_weights_strs) == 0:
        unet_block_weights = [None]
    else:
        unet_block_weights = [float(f)
                          for s in unet_block_weights_strs
                          for f in s.split(',')]

    merge_configs = []
    for a in merge_algorithms:
        for ma in merge_alphas:
            for ua in unet_alphas:
                for ta in text_encoder_alphas:
                    for bw in unet_block_weights:
                        merge_configs.append({
                            'alpha': ma,
                            'algorithm': a,
                            'unet_alpha': ua,
                            'text_encoder_alpha': ta,
                            'block_weights': bw,
                        })
    return merge_configs




def render_all(prompts: list[str], negative_prompts: Optional[list[str]], seeds: list[int], cfg: float,
               repo_ids_or_paths: list[str],
               device: str,
               size: tuple[int, int],
               batch_size: int,
               inference_steps: int = 15,
               save_partial_filename: Optional[str] = None,
               local_files_only: bool = False,
               merge_config: Optional[dict] = None,
               disable_nsfw_checker: bool = False,
               ) -> Image:
    all_images = []
    print(f"{len(prompts)} prompts")

    row_labels = []

    def save_partial_if_requested():
        if save_partial_filename is not None:
            num_rows = int(len(all_images) / len(prompts))
            grid_image = make_image_grid(all_images, num_rows=num_rows, num_cols=len(prompts),
                                         row_labels=row_labels[:num_rows], col_labels=prompts)
            grid_image.save(save_partial_filename)

    if merge_config is not None:
        if len(repo_ids_or_paths) != 2 and len(repo_ids_or_paths) != 3:
            raise ValueError("Must specify either 2 or 3 models when using merge_alphas")

        merge_configs = explode_merge_config(merge_config)

        for merge_index, config in enumerate(tqdm(merge_configs)):
            unet_block_weights = config['block_weights']
            if unet_block_weights is not None and len(unet_block_weights) != 25:
                raise ValueError(f"Wrong number of unet block weights. There must be 25, eg \"0.0,0.04,0.08,0.12,0.16,"
                                 f"0.2,0.24,0.28,0.32,0.36,0.4,0.44,0.48,0.52,0.56,0.6,0.64,0.68,0.72,0.76,0.8,0.84,0.88,"
                                 f"0.92,0.96\". You have specified \"{unet_block_weights}\", which is not valid.")

            merge_algorithm = config['algorithm']
            per_module_alphas = {
                'unet': config['unet_alpha'],
                'text_encoder': config['text_encoder_alpha']
            }
            alpha = config['alpha']
            this_model_label = f"{merge_algorithm} merge of {repo_ids_or_paths} with alpha {alpha}, " \
                               f"per_module_alphas {per_module_alphas}, " \
                               f"unet_block_weights {unet_block_weights}"
            row_labels.append(this_model_label)
            print(this_model_label)
            model_c = repo_ids_or_paths[2] if len(repo_ids_or_paths) == 3 else None
            pipeline = merge_models(repo_ids_or_paths[0], repo_ids_or_paths[1], model_c, alpha=alpha,
                                    algorithm=merge_algorithm, unet_block_weights=unet_block_weights,
                                    per_module_alphas=per_module_alphas,
                                    local_files_only=local_files_only)
            row_images = render_row(prompts,
                                    negative_prompts=negative_prompts,
                                    seeds=seeds,
                                    pipeline=pipeline,
                                    device=device,
                                    batch_size=batch_size,
                                    cfg=cfg,
                                    num_inference_steps=inference_steps,
                                    sample_w=size[0], sample_h=size[1],
                                    disable_nsfw_checker=disable_nsfw_checker,)
            all_images += row_images
            save_partial_if_requested()

            save_merge_path_prefix = merge_config.get('save_merge_path_prefix', None)
            if save_merge_path_prefix is not None:
                save_half = merge_config.get('save_merge_half', True)
                if save_half:
                    pipeline.to(torch.float16)
                save_merge_path = f"{save_merge_path_prefix}_{merge_index:02d}"
                pipeline.save_pretrained(save_merge_path)

            del pipeline

        grid_image = make_image_grid(all_images, len(merge_configs), len(prompts), row_labels, prompts)
    else:
        row_labels = repo_ids_or_paths
        for repo_id_or_path in tqdm(repo_ids_or_paths):
            print(f"model {repo_id_or_path}:")
            pipeline = load_model(repo_id_or_path, local_files_only=local_files_only)
            row_images = render_row(prompts,
                                    negative_prompts=negative_prompts,
                                    seeds=seeds,
                                    pipeline=pipeline,
                                    device=device,
                                    batch_size=batch_size,
                                    cfg=cfg,
                                    num_inference_steps=inference_steps,
                                    sample_w=size[0], sample_h=size[1],
                                    disable_nsfw_checker=disable_nsfw_checker,)
            all_images += row_images
            save_partial_if_requested()
            del pipeline
        grid_image = make_image_grid(all_images, len(repo_ids_or_paths), len(prompts), row_labels, prompts)

    return grid_image


def main():
    parser = argparse.ArgumentParser(
        prog="grate",
        description="Generates a grid of images by running a set of prompts through different Stable Diffusion models.",
    )

    parser.add_argument("--prompts",
                        nargs='+',
                        required=True,
                        help=(
                                    "EITHER: a path to a JSON file containing prompt and negative prompt pairs eg [{'prompt': 'a fish', 'negative_prompt': 'distorted', 'seed': 123}, ...]. \n\n" +
                                    "OR: multiple strings enclosed in \"\" and separated by spaces. eg --prompts \"a cat\" \"a dog\" \"a fish\"")
                        )
    parser.add_argument("--repo_ids_or_paths",
                        nargs='+',
                        required=True,
                        help="repository ids or paths to models in diffusers or ckpt format, as strings enclosed in \"\" "
                             "and separated by spaces. eg "
                             "--repo_ids_or_paths \"stablityai/stable-diffusion-2-1\" \"../models/v1-5-pruned-emaonly.ckpt\"")
    parser.add_argument("--output_path",
                        required=True,
                        help="Where to save the resulting image. Also saves partially-rendered images to this location as each row finishes rendering.")
    parser.add_argument("--device",
                        required=False,
                        default=None,
                        help="(Optional) Device to use, eg 'cuda', 'mps', 'cpu'. if omitted, will try to pick the best device.")
    parser.add_argument("--batch_size",
                        required=False,
                        type=int,
                        default=1,
                        help="(Optional, default=1) Batch size.")
    parser.add_argument("--width",
                        required=False,
                        default=512,
                        type=int,
                        help="(Optional, default=512) Individual image width")
    parser.add_argument("--height",
                        required=False,
                        type=int,
                        default=512,
                        help="(Optional, defatul=512) Individual image height")
    parser.add_argument("--negative_prompts",
                        required=False,
                        type=str,
                        nargs="+",
                        help="(Optional) Negative prompts. Specify either one string to share for all `--prompts`, or as many strings as there are `--prompts`.")
    parser.add_argument("--seeds",
                        required=False,
                        type=int,
                        nargs="+",
                        help="(Optional) Seeds. Specify either one seed to share for all `--prompts`, or as many seeds as there are `--prompts`.")
    parser.add_argument("--cfg",
                        required=False,
                        type=float,
                        default=7.5,
                        help="(Optional, default=7.5) CFG scale.")
    parser.add_argument("--steps",
                        required=False,
                        type=int,
                        default=15,
                        help="(Optional, default=15) How many inference steps to run")
    parser.add_argument("--disable_nsfw_checker",
                        required=False,
                        action="store_true",
                        help="(Optional)")
    parser.add_argument("--local_files_only",
                        required=False,
                        action='store_true',
                        help="(Optional) Use only local data (do not attempt to download or update models)")
    parser.add_argument("--merge_alpha",
                        required=False,
                        type=float,
                        nargs="+",
                        help="(Optional) If set, --repo_ids_or_paths must specify either 2 or 3 models, which will be merged using the given alpha and used in place of multiple models.")
    parser.add_argument("--merge_algorithm",
                        required=False,
                        type=str,
                        default=None,
                        nargs="+",
                        help="(Optional) If doing merges, the algorithm to use - one of 'weighted_sum', 'sigmoid', 'inv_sigmoid', or 'add_diff'. 'add_diff' only works for 3-way merge. ")
    parser.add_argument("--merge_unet_block_weights",
                        required=False,
                        type=str,
                        nargs="+",
                        help="(Optional) 25 comma-separated floats specified as strings, eg \"0.0, 0.0, 0.0, (... 22 more floats)\", to merge each part of the UNet using a different weight ('block-weighted merging').")
    parser.add_argument("--merge_unet_alpha",
                        required=False,
                        type=float,
                        nargs="+",
                        help="(Optional) Override the merge alpha with a unet-specific alpha.")
    parser.add_argument("--merge_text_encoder_alpha",
                        required=False,
                        type=float,
                        nargs="+",
                        help="(Optional) Override the merge alpha with a text-encoder-specific alpha.")
    parser.add_argument("--save_merge_path_prefix",
                        required=False,
                        type=str,
                        help="(Optional) If doing a merge, save all merge combinations using this path as a prefix.")
    parser.add_argument("--save_merge_float32",
                        action="store_true",
                        help="(Optional) If saving merges, save with float32 precision (default is float16).")
    args = parser.parse_args()


    def use_arg_list_or_expand_or_default(arg: list, required_count: int, default: list|None) -> list:
        if arg is None:
            return default
        elif len(arg) == 1:
            # expand to required count
            return arg * required_count
        else:
            return arg


    if len(args.prompts) == 1 and os.path.isfile(args.prompts[0]):
        with open(args.prompts[0], 'rt') as f:
            prompts_json = json.load(f)
            prompts = [p.get('prompt', '') for p in prompts_json]
            negative_prompts = [p.get('negative_prompt', '') for p in prompts_json]
            seeds = [int(p.get('seed', 1 + i)) for i, p in enumerate(prompts_json)]
            print(f"loaded {len(prompts)} prompts from {args.prompts[0]}")
            print({'prompts': prompts, 'negative_prompts': negative_prompts, 'seeds': seeds})
    else:

        prompts = args.prompts
        negative_prompts = use_arg_list_or_expand_or_default(args.negative_prompts, len(prompts), [''] * len(prompts))
        seeds = use_arg_list_or_expand_or_default(args.seeds, len(prompts), [1 + i for i in range(len(prompts))])

    print(f"args.merge_unet_alpha: {args.merge_unet_alpha}")
    merge_config = {
        'alphas': args.merge_alpha,
        'algorithms': args.merge_algorithm,
        'unet_alphas': args.merge_unet_alpha,
        'text_encoder_alphas': args.merge_text_encoder_alpha,
        'block_weights': args.merge_unet_block_weights,
    }
    if all(v is None for v in merge_config.values()):
        merge_config = None
    else:
        merge_config['save_merge_path_prefix'] = args.save_merge_path_prefix
        merge_config['save_merge_half'] = not args.save_merge_float32


    render_all(prompts=prompts,
               negative_prompts=negative_prompts,
               seeds=seeds,
               repo_ids_or_paths=args.repo_ids_or_paths,
               merge_config=merge_config,
               device=args.device,
               size=(args.width, args.height),
               batch_size=args.batch_size,
               cfg=args.cfg,
               inference_steps=args.steps,
               local_files_only=args.local_files_only,
               save_partial_filename=args.output_path,
               disable_nsfw_checker=args.disable_nsfw_checker,)
    print(f"grate saved to {args.output_path}")

if __name__ == '__main__':
    main()
